# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: angr/protos/cfg.proto
# Protobuf Python Version: 6.33.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    33,
    0,
    '',
    'angr/protos/cfg.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from angr.protos import primitives_pb2 as angr_dot_protos_dot_primitives__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15\x61ngr/protos/cfg.proto\x12\x0b\x61ngr.protos\x1a\x1c\x61ngr/protos/primitives.proto\"]\n\x07\x43\x46GNode\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x10\n\x08\x62lock_id\x18\x03 \x03(\x03\x12\x11\n\treturning\x18\x04 \x01(\x08\x12\x13\n\x0binstr_addrs\x18\x05 \x03(\x04\"\xcd\x01\n\x03\x43\x46G\x12\r\n\x05ident\x18\x01 \x01(\t\x12#\n\x05nodes\x18\x02 \x03(\x0b\x32\x14.angr.protos.CFGNode\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.angr.protos.Edge\x12,\n\x0bmemory_data\x18\x04 \x03(\x0b\x32\x17.angr.protos.MemoryData\x12\x12\n\nnormalized\x18\x05 \x01(\x08\x12.\n\x0bjump_tables\x18\x06 \x03(\x0b\x32\x19.angr.protos.IndirectJump\"\xfb\x02\n\nMemoryData\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x11\n\x04size\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x34\n\x04type\x18\x03 \x01(\x0e\x32&.angr.protos.MemoryData.MemoryDataType\x12\x1b\n\x0ereference_size\x18\x04 \x01(\rH\x01\x88\x01\x01\"\xde\x01\n\x0eMemoryDataType\x12\x13\n\x0fUnknownDataType\x10\x00\x12\x0f\n\x0bUnspecified\x10\x01\x12\x0b\n\x07Integer\x10\x02\x12\x10\n\x0cPointerArray\x10\x03\x12\n\n\x06String\x10\x04\x12\x11\n\rUnicodeString\x10\x05\x12\x13\n\x0fSegmentBoundary\x10\x06\x12\x11\n\rCodeReference\x10\x07\x12\x0f\n\x0bGOTPLTEntry\x10\x08\x12\r\n\tELFHeader\x10\t\x12\x11\n\rFloatingPoint\x10\n\x12\r\n\tAlignment\x10\x0b\x42\x07\n\x05_sizeB\x11\n\x0f_reference_size\"^\n\rJumptableInfo\x12\x11\n\x04\x61\x64\x64r\x18\x01 \x01(\x04H\x00\x88\x01\x01\x12\x0c\n\x04size\x18\x02 \x01(\x05\x12\x12\n\nentry_size\x18\x03 \x01(\x05\x12\x0f\n\x07\x65ntries\x18\x04 \x03(\x04\x42\x07\n\x05_addr\"\xd0\x01\n\x0cIndirectJump\x12\x0c\n\x04\x61\x64\x64r\x18\x01 \x01(\x04\x12\x10\n\x08ins_addr\x18\x02 \x01(\x04\x12\x11\n\tfunc_addr\x18\x03 \x01(\x04\x12\x10\n\x08jumpkind\x18\x04 \x01(\t\x12\x10\n\x08stmt_idx\x18\x05 \x01(\x05\x12\x18\n\x10resolved_targets\x18\x06 \x03(\x04\x12\x11\n\tjumptable\x18\x07 \x01(\x08\x12.\n\njumptables\x18\x08 \x03(\x0b\x32\x1a.angr.protos.JumptableInfo\x12\x0c\n\x04type\x18\t \x01(\x05\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'angr.protos.cfg_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CFGNODE']._serialized_start=68
  _globals['_CFGNODE']._serialized_end=161
  _globals['_CFG']._serialized_start=164
  _globals['_CFG']._serialized_end=369
  _globals['_MEMORYDATA']._serialized_start=372
  _globals['_MEMORYDATA']._serialized_end=751
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_start=501
  _globals['_MEMORYDATA_MEMORYDATATYPE']._serialized_end=723
  _globals['_JUMPTABLEINFO']._serialized_start=753
  _globals['_JUMPTABLEINFO']._serialized_end=847
  _globals['_INDIRECTJUMP']._serialized_start=850
  _globals['_INDIRECTJUMP']._serialized_end=1058
# @@protoc_insertion_point(module_scope)
